// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawnlibs.hpp"

namespace Dawn {
  template<typename ...A>
  class Event {
    private:
      int32_t nextId = 0;
      std::map<int32_t, std::function<void(A...)>> listeners;

    public:
      /**
       * Constructs a new Event.
       */
      Event() {

      }

      /**
       * Emits the event.
       * @param args The arguments to pass to the listeners.
       */
      void emit(A ...args) {
        auto copy = listeners;
        for(auto &pair : copy) {
          pair.second(args...);
        }
      }

      /**
       * Listens to the event.
       * @param listener The listener to add.
       * @returns A function that can be called to remove the listener.
       */
      std::function<void()> listen(
        const std::function<void(A...)> listener
      ) {
        int32_t id = nextId++;
        listeners[id] = listener;
        return [this, id]() {
          listeners.erase(id);
        };
      }

      /**
       * Destroys the event.
       */
      virtual ~Event() {
        listeners.clear();
      }
  };
}