// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawnlibs.hpp"

#define SCENE_COMPONENT_STATE_INIT 0x01
#define SCENE_COMPONENT_STATE_DISPOSED 0x02

namespace Dawn {
  class Game;
  class Scene;
  class SceneItem;

  class SceneComponent : std::enable_shared_from_this<SceneComponent> {
    private:
      std::weak_ptr<SceneItem> item;
      uint_fast8_t sceneComponentState = 0;

    protected:
      /**
       * Custom component listener that is invoked when the component is meant
       * to initialize.
       */
      virtual void onInit() = 0;

      /**
       * Custom component listener that is invoked when the component is meant
       * to dispose.
       */
      virtual void onDispose() = 0;

    public:
      /**
       * Initializes this scene component.
       * 
       * @param item Scene item that this component belongs to.
       */
      void init(const std::shared_ptr<SceneItem> item);

      /**
       * Disposes this scene component.
       */
      void dispose();

      /**
       * Returns the scene item that this scene component belongs to.
       * 
       * @return Reference to the scene item that this component belongs to.
       */
      std::shared_ptr<SceneItem> getItem();

      /**
       * Returns the scene that this scene component belongs to.
       * 
       * @return Reference to the scene that this component belongs to.
       */
      std::shared_ptr<Scene> getScene();

      /**
       * Returns the game that this scene component belongs to.
       * 
       * @return Reference to the game that this component belongs to.
       */
      std::shared_ptr<Game> getGame();

      /**
       * Disposes this scene component.
       */
      virtual ~SceneComponent();
  };
}