// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawnlibs.hpp"

namespace Dawn {
  class SceneItem;

  class SceneItemTransform {
    protected:
      // Local (real) values
      glm::vec3 localPosition;
      glm::vec3 localScale;
      glm::quat localRotation;

      // Cached (non-real) values
      glm::mat4 transformLocal;
      glm::mat4 transformWorld;

      // Heirarchy
      std::weak_ptr<SceneItem> parent;
      std::vector<std::weak_ptr<SceneItem>> children;

      /**
       * Unpacks the local transform into the local transform matrix values.
       */
      void updateLocalValuesFromLocalTransform();

      /**
       * Sets the local transform (and all local real values) to match the world
       * transform values.
       */
      void updateLocalTransformFromWorldTransform();

      /**
       * Updates the world transform of this items children.
       */
      void updateChildrenWorldTransforms();

      /**
       * Called by the parent of this item when it's transform changes.
       */
      void updateWorldTransformFromParent();

      /**
       * Updates the local transform of this item from the local values.
       */
      void updateLocalTransformFromLocalValues();

      /**
       * Updates the world transform from this items local transform.
       */
      void updateWorldTransformFromLocalTransform();

    public:
      SceneItemTransform();

      /**
       * Returns the parent of this scene item.
       * 
       * @return Pointer to the parent of this scene item.
       */
      std::shared_ptr<SceneItem> getParent();

      /**
       * Returns the children of this scene item.
       * 
       * @return Vector of pointers to the children of this scene item.
       */
      std::vector<std::shared_ptr<SceneItem>> getChildren();

      /**
       * Returns the local transform of this item (relative to its parent).
       * 
       * @return Local transform of this item.
       */
      glm::mat4 getLocalTransform();

      /**
       * Returns the world transform of this item (relative to scene root).
       * 
       * @return World transform of this item.
       */
      glm::mat4 getWorldTransform();

      /**
       * Gets the local position of this item (relative to its parent).
       * 
       * @return Local position of this item.
       */
      glm::vec3 getLocalPosition();

      /**
       * Local scale of this item (relative to its parent).
       * 
       * @return Local scale of this item.
       */
      glm::vec3 getLocalScale();

      /**
       * Local rotation of this item (relative to its parent).
       * 
       * @return Local rotation of this item.
       */
      glm::quat getLocalRotation();

      /**
       * Sets the transform of this item within local space (relative to parent)
       * 
       * @param transform Transform of this item in local space.
       */
      void setLocalTransform(const glm::mat4 transform);
      
      /**
       * Sets the transform of this item within world space (relative to scene
       * root)
       * 
       * @param transform Transform of this item in world space.
       */
      void setWorldTransform(const glm::mat4 transform);

      /**
       * Sets the local position of this item (relative to its parent).
       * 
       * @param position Local position of this item.
       */
      void setLocalPosition(const glm::vec3 position);

      /**
       * Sets the local scale of this item (relative to its parent).
       * 
       * @param scale Local scale of this item.
       */
      void setLocalScale(const glm::vec3 scale);

      /**
       * Sets the local rotation of this item (relative to its parent).
       * 
       * @param rotation Local rotation of this item.
       */
      void setLocalRotation(const glm::quat rotation);

      /**
       * Returns whether or not this item is a child of the given item. This
       * will traverse up the entire heirarchy to confirm.
       * 
       * @return True if this item is within the given item's heirarchy.
       */
      bool_t isChildOf(std::shared_ptr<SceneItem> item);

      /**
       * Orients this transform to look at a given point in world space.
       * 
       * @param position Position of the origin of this transform.
       * @param look Position in world space this transform looks at.
       * @param up Up vector of this transform.
       */
      void lookAt(
        const glm::vec3 position,
        const glm::vec3 look,
        const glm::vec3 up
      );

      virtual ~SceneItemTransform();
  };
}